# SPDX-FileCopyrightText: 2020 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime
import os
from typing import List, Optional

from pycdstar3 import CDStar, CDStarVault
from pycdstar3.api import CDStarArchive

from uploader import config
from uploader import models_cdstar as cdmodels


def upload_files(filelist: List[str], mime_type: Optional[str]) -> cdmodels.ArchiveInfo:
    """Uploads a list of files to CDSTAR.

    For each filepath in filelist, a call to `upload_file_to_archive` is made.

    Parameters
    ----------
    filelist : List[str]
        A list of strings, where each sting is a full path to the file which shall be uploaded to CDSTAR.
    mime_type: str
        String representation of a MIME type that shall be assigned to all uploaded files. If `None`, it
        will use the auto-detection method of CDSTAR to determine the MIME type.

    Returns
    -------
    List[models.ArchiveInfo]
        A list of Archives modelled after [CDSTARs ArchiveInfo data structure](https://cdstar.gwdg.de/docs/dev/#ArchiveInfo).
    """

    cdstar = CDStar(
        config.BasicSettings().cdstar_uri,
        auth=(config.BasicSettings().cdstar_user, config.BasicSettings().cdstar_pass),
    )

    with cdstar.begin(autocommit=True):
        vault = CDStarVault(cdstar, config.BasicSettings().cdstar_vault_target)
        archive = vault.new_archive()
        response = cdmodels.ArchiveInfo(
            id=archive.id,
            vault=vault.name,
            created=datetime.datetime.now().isoformat(),
            modified=datetime.datetime.now().isoformat(),
            file_count=0,
        )

        for file in filelist:
            uploaded = upload_file_to_archive(file, archive, mime_type)
            response.file_count += 1
            response.files.append(uploaded)
        response.modified = datetime.datetime.now()

    return response


def upload_file_to_archive(file: str, archive: CDStarArchive, mime_type: Optional[str]) -> cdmodels.FileInfo:
    """Uploads a file to a CDStarArchive.

    If not CDSTAR archive is supplied, a new archive will be created prior to uploading a file.

    Parameters
    ----------
    file : str
        A full path to a file that shall be uploaded to the CDSTAR archive.
    archive : CDStarArchive
        An instance of a CDStarArchive. If None, it will be created prior to uploading file.
    mime_type : str
        The MIME-Type of the to-be-uploaded file. If None, the automatic detection of CDSTAR will be used.

    Returns
    -------
    models.FileInfo
        The result of the upload procedure modelled after [CDSTARs FileInfo data structure](https://cdstar.gwdg.de/docs/dev/#FileInfo).
        On error, the ID of the annotation will be `None`.
    """

    if os.path.isfile(file) and os.path.exists(file) and os.access(file, os.R_OK):
        filename = os.path.basename(file)
        cdstar_file = archive.file(filename)
        kargs = {"type": None, "source": file}
        if mime_type is not None:
            kargs["type"] = mime_type
        response = cdstar_file.put(**kargs)
        if "id" not in response.keys() or "name" not in response.keys() or "type" not in response.keys():
            return cdmodels.FileInfo(id=None, name=filename)
        return cdmodels.FileInfo.parse_obj(response)
    else:
        return cdmodels.FileInfo(id=None, name=file)
