# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import fastapi
import fastapi.responses
from fastapi import BackgroundTasks, FastAPI, HTTPException, Request, Depends
from sqlalchemy.orm import Session

import datetime
import logging
import os
from typing import List

from uploader import cdstar
from uploader.config import BasicSettings
from uploader import errors
from uploader import models_activeworkflow as awmodels
from uploader import models_internal as intmodels
from uploader import store
from uploader import utils

app = FastAPI()
access_log = logging.getLogger("gunicorn.access")
error_log = logging.getLogger("gunicorn.error")


@app.get("/appinfo")
def appinfo():
    """Prints a part of the configuration of this datalake-agent instance."""

    access_log.info(BasicSettings())
    return BasicSettings().dict(exclude={"cdstar_pass", "cdstar_user", "db_uri"})


@app.post(
    "/datalake",
    response_model=awmodels.ResponseCommon,
    response_model_exclude_none=True,
)
def datalake(
    payload: awmodels.RequestCommon,
    background_tasks: BackgroundTasks,
    db: Session = Depends(store.get_db),
):
    """Implements an remote agent API endpoint as specified by active_workflow, see
    [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md) for more details.
    See the explicit API endpoints for each method to find more information on them.
    """

    if payload.method is not None and payload.method == "register":
        return register(awmodels.RequestRegister.parse_obj(payload.dict()), background_tasks)
    elif payload.method is not None and payload.method == "receive":
        return receive(awmodels.RequestReceive.parse_obj(payload.dict()), background_tasks, db)
    elif payload.method is not None and payload.method == "check":
        return check(awmodels.RequestCheck.parse_obj(payload.dict()), background_tasks, db)

    raise HTTPException(
        status_code=400,
        detail="You need to supply a valid method: register, receive, check.",
    )


@app.post(
    "/datalake/register",
    response_model=awmodels.ResponseRegister,
    response_model_exclude_none=True,
    status_code=fastapi.status.HTTP_200_OK,
)
def register(payload: awmodels.RequestRegister, background_tasks: BackgroundTasks):
    """Handle register method of active_workflow. Check
    [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md#register-method) for
    more information on the method itself. The response from this method will provide some examples for annotations that may be sent.
    """

    access_log.info(f"Register - {datetime.datetime.now().isoformat()} - {payload.dict()}")

    return awmodels.ResponseRegister()


@app.post(
    "/datalake/receive",
    response_model=awmodels.ResponseReceive,
    response_model_exclude_none=True,
)
def receive(
    payload: awmodels.RequestReceive,
    background_tasks: BackgroundTasks,
    db: Session = Depends(store.get_db),
):
    """Starts and upload and annotation of a list of files.

    Handle the receive method of ActiveWorkflow, see
    [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md#receive-method) for more information.
    This method checks `$.params.message.payload.files_moved` and expects a list of filenames there. It will iterate through each of these filenames
    and upload them to a new CDSTAR Archive per message. The filenames are expected to reside under
    `base_directory`/`source_directory` as configured (or overriden in the agent options in active workflow.)
    Metadata annotations supplied under `$.params.message.payload.annotations` or
    (with lower priority) under `$.params.options.annotations` are added to generated metadata (from CDSTAR responses)
    and uploaded to CouchDB. Annotations starting with `@` will be ignored, they will be set by the application where applicable.
    """

    access_log.info(f"Receive - {datetime.datetime.now().isoformat()} - {payload.dict()}")

    if not utils.is_authorized(payload.params):
        raise errors.UnauthorizedException(
            message='Please supply valid credentials using the `api_key` option parameter referring to a stored credentials key. You may use Liquid templating: `"api_key": "{% credential credential_name %}"`.'
        )

    response = awmodels.ResponseReceive()
    response.result.memory = payload.params.memory

    opts = payload.params.options.dict()
    for k, v in payload.params.message.payload.options.dict(
        exclude_none=True, exclude_defaults=True, exclude_unset=True
    ).items():
        opts[k] = v

    msg_payload = payload.params.message.payload
    if msg_payload.files_moved is None or len(msg_payload.files_moved) < 1:
        error_log.debug(f"Received message specifies no files moved. Using filemask to find files.")
        filelist = utils.retrieve_files_by_mask(subdir=opts["source_directory"], filemask=opts["source_filemask"])
    else:
        error_log.debug(f"Received message specifies files. Using namelist to find files.")
        filelist = utils.retrieve_files_by_name(subdir=opts["source_directory"], namelist=msg_payload.files_moved)

    upload_procedure = store.start_upload_procedure(db)
    upload_id = upload_procedure.id
    background_tasks.add_task(
        perform_datalake_upload,
        filelist=filelist,
        options=opts,
        upload_id=upload_id,
        db=db,
    )

    response.result.logs.append(f"The upload has been started with id {upload_id}")
    response.result.memory.uploads.append(upload_id)
    error_log.debug(f"The upload has been started with id {upload_id}")
    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_202_ACCEPTED,
        content=response.dict(exclude_none=True),
    )


@app.post(
    "/datalake/check",
    response_model=awmodels.ResponseCheck,
    response_model_exclude_none=True,
    status_code=fastapi.status.HTTP_200_OK,
)
def check(
    payload: awmodels.RequestCheck,
    background_tasks: BackgroundTasks,
    db: Session = Depends(store.get_db),
):
    """Check on data uploads as supplied by `$.params.memory.uploads`.

    Handles the check method of ActiveWorkflow, see
    [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md#check-method) for more information.
    This method uses upload ids supplied by the memory part of a request to check on all uploads. An upload may be running or completed,
    which will be reported in the logs or message section respectively. Invalid / unknown upload ids are removed from the memory.
    """

    access_log.info(f"Check - {datetime.datetime.now().isoformat()} - {payload.dict()}")

    if not utils.is_authorized(payload.params):
        raise errors.UnauthorizedException(
            message='Please supply valid credentials using the `api_key` option parameter referring to a stored credentials key. You may use Liquid templating: `"api_key": "{% credential credential_name %}"`.'
        )

    response = awmodels.ResponseCheck()

    running_uploads = {upl.id: upl for upl in store.get_uploads_by_state(db, intmodels.UploadState.started)}
    error_log.debug(f"Running Uploads: {running_uploads.keys()}")

    completed_uploads = {upl.id: upl for upl in store.get_uploads_by_state(db, intmodels.UploadState.completed)}
    error_log.debug(f"Completed Uploads: {completed_uploads.keys()}")

    for upload_id in payload.params.memory.uploads:
        if upload_id in running_uploads.keys():
            response.result.memory.uploads.append(upload_id)
            response.result.logs.append(
                f"Upload {upload_id} is running since {running_uploads[upload_id].started_at.isoformat()}"
            )
        elif upload_id in completed_uploads.keys():
            upl = completed_uploads[upload_id]
            upl.state = intmodels.UploadState.retrieved
            upl.retrieved_at = datetime.datetime.now()
            upl = store.store_upload(db, upl)
            upl_response = awmodels.ResponseCheck.parse_obj(upl.response)
            response.result.logs = response.result.logs + upl_response.result.logs
            response.result.errors = response.result.errors + upl_response.result.errors
            response.result.messages = response.result.messages + upl_response.result.messages
        else:
            error_log.warning(f"Found orphan upload ID: {upload_id}")

    return response


def perform_datalake_upload(filelist: List[str], options: dict, upload_id: int, db: Session) -> None:
    """Perform the upload and annotation within the datalake.
    From the options and annotations supplied by the caller or configured in the application, find all files
    using `source_directory` and `source_filemask` from the `options` parameter. Upload each file to a new
    archive created in CDSTAR and create metadata annotations supplied by the `annotations` parameter. Each key
    in annotations will be filtered against the internal models of `models_metadata`.

    Parameters
    ----------
    options : dict
        A dictionary of options that are supplied by the caller. Specifically, `source_directory`, `source_filemask` and
        `delete_after_upload` have to be set.
    annotations : dict
        A dictionary of annotations that are added to the CouchDB metadata after the files have been uploaded to CDSTAR.
    """
    response = awmodels.ResponseCheck()

    if filelist is None or len(filelist) < 1:
        response.result.logs.append(f"No files found to upload for id {upload_id}.")
        error_log.debug(f"No files found to upload for id {upload_id}.")
        upload_completed(upload_id, response, db)
        return

    # if specified, create TAR archive before uploading files
    if options["tar_before_upload"]:
        filelist = utils.tar_before_upload(files=filelist)

    # upload all files to a new cdstar archive
    try:
        # time_format should produce e.g. this: "2021-03-11T16:56:18"
        time_format = "%Y-%m-%dT%H:%M:%S"
        start_time = datetime.datetime.now().strftime(time_format)

        uploaded = cdstar.upload_files(filelist, mime_type=options["mime_type"])

        end_time = datetime.datetime.now().strftime(time_format)
        response.result.messages.append(
            awmodels.MessageOutput(
                upload_id=upload_id, archive_id=uploaded.id, start_time=start_time, end_time=end_time
            )
        )
        response.result.logs.append(
            f"Upload to CDSTAR for ID {upload_id} finished at {datetime.datetime.now().isoformat()}."
        )
        error_log.debug(
            f"Upload to CDSTAR for ID {upload_id} (CDSTAR Archive: {uploaded.id} finished at {datetime.datetime.now().isoformat()}."
        )
    except:
        response.result.errors.append(
            f"There was an error while uploading to CDSTAR. Check the server error log for more information. Upload with ID {upload_id} failed."
        )
        # TODO: debug error_log.debug(traceback.format_exc())
        error_log.error(
            f"There was an error while uploading to CDSTAR. Check the server error log for more information. Upload with ID {upload_id} failed."
        )
        upload_completed(upload_id, response, db)
        return

    response.result.logs.append(uploaded.json(exclude_none=True, exclude_defaults=True, exclude_unset=True))
    for file in uploaded.files:
        if file.id is not None:
            if options["delete_after_upload"]:
                os.remove(
                    os.path.join(
                        BasicSettings().base_directory,
                        options["source_directory"].lstrip("/"),
                        file.name.lstrip("/"),
                    )
                )
        else:
            response.result.errors.append(f"Error: {file.name} could not be uploaded.")
            error_log.debug(f"Error: {file.name} could not be uploaded.")

    upload_completed(upload_id, response, db)


def upload_completed(upload_id: int, response: awmodels.ResponseCheck, db: Session):
    upload = store.get_upload_by_id(db, upload_id)
    if upload == None:
        error_log.error(f"I was not able to find ID {upload_id} which has response {response.dict(exclude_none=True)}")
        return
    upload.state = intmodels.UploadState.completed
    upload.completed_at = datetime.datetime.now()
    upload.response = response.dict(exclude_none=True)
    store.store_upload(db, upload)
    error_log.info(f"Completed upload of {upload_id}.")


@app.exception_handler(errors.ActiveWorkflowException)
async def aw_exception_handler(request: Request, exc: errors.ActiveWorkflowException):
    error_log.error(f"{exc.name}: {exc.message}")
    resp = awmodels.ResponseCheck()
    resp.result.errors.append(f"{exc.name}: {exc.message}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_400_BAD_REQUEST,
        content=resp.dict(exclude_none=True),
    )


@app.exception_handler(errors.InternalException)
async def int_exception_handler(request: Request, exc: errors.InternalException):
    error_log.error(f"{exc.name}: {exc.message}")
    resp = awmodels.ResponseCheck()
    resp.result.errors.append(f"{exc.name}: {exc.message}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR,
        content=resp.dict(exclude_none=True),
    )


@app.exception_handler(errors.UnauthorizedException)
async def aw_unauthorized_handler(request: Request, exc: errors.UnauthorizedException):
    error_log.info(f"Unauthorized access - {exc.name}: {exc.message}")
    resp = awmodels.ResponseCheck()
    resp.result.errors.append(f"{exc.name}: {exc.message}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_401_UNAUTHORIZED,
        content=resp.dict(exclude_none=True),
    )
