# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import abc
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, validator

from uploader import config


#
# Common models
#
class CredentialsCommon(BaseModel):
    """
    This model represents the expected credentials content to communicate credentials with the agent.
    """

    name: str = Field("", example="")
    value: str = Field("", example="")

    @staticmethod
    def example() -> dict:
        return {"name": "", "value": ""}


class ParamsCommon(BaseModel, abc.ABC):
    """
    These are all possible parameters that may be supplied using the Active Workflow Remote Agent API.
    """

    message: Optional[Dict[str, Any]] = {}
    options: Optional[Dict[str, Any]] = {}
    memory: Optional[Dict[str, Any]] = {}
    credentials: Optional[List[CredentialsCommon]] = []


class ResultCommon(BaseModel, abc.ABC):
    """
    These are all possible result values that may be returned using the Active Workflow Remote Agent API.
    """

    errors: Optional[List[str]] = None
    logs: Optional[List[str]] = None
    memory: Optional[Dict[str, Any]] = None
    messages: Optional[List[Dict[str, Any]]] = None
    name: Optional[str] = None
    display_name: Optional[str] = None
    description: Optional[str] = None
    default_options: Optional[Dict[str, Any]] = None


class RequestCommon(BaseModel, abc.ABC):
    """
    These are all possible values in a request from Active Workflow using the Remote Agent API.
    """

    method: str = Field(..., example="receive")
    params: ParamsCommon = Field(..., example=ParamsCommon())

    @validator("method")
    def method_choices(cls, v):
        if v not in ["register", "receive", "check"]:
            raise ValueError("Method has to be one of 'register', 'receive' or 'check'.")
        return v


class ResponseCommon(BaseModel, abc.ABC):
    """
    This is a possible reponse returned when using the Acitve Workflow Remote Agent API.
    """

    result: ResultCommon = Field(..., example=ResultCommon())


#
# Application specifics
#


class MessageOutput(BaseModel):
    """
    This model represents the produced output schema that will be provided in the messages.
    """

    upload_id: int = Field(..., example=4)
    archive_id: str = Field(..., example="iw7s2px9")
    start_time: str = Field(..., example="2021-03-11T16:56:18")
    end_time: str = Field(..., example="2021-03-11T16:56:20")

    @staticmethod
    def example() -> dict:
        return {
            "upload_id": 4,
            "archive_id": "iw7s2px9",
            "start_time": "2021-03-11T16:56:18",
            "end_time": "2021-03-11T16:56:20",
        }


class OptionsCommon(BaseModel):
    """
    This model represents expected options to configure this agent (or rather a call to this agent).
    """

    source_directory: str = Field(config.BasicSettings().source_directory)
    source_filemask: str = Field(config.BasicSettings().source_filemask)
    mime_type: str = Field(config.BasicSettings().mime_type)
    delete_after_upload: bool = Field(config.BasicSettings().delete_after_upload)
    tar_before_upload: bool = Field(config.BasicSettings().tar_before_upload)
    api_key: str = Field("")

    @staticmethod
    def example() -> Dict[str, Any]:
        return {
            "source_directory": config.BasicSettings().source_directory,
            "source_filemask": config.BasicSettings().source_filemask,
            "mime_type": config.BasicSettings().mime_type,
            "delete_after_upload": config.BasicSettings().delete_after_upload,
            "api_key": "",
        }


class PayloadInput(BaseModel):
    """
    This model represents the expected input payload when a message is received via `receive`.
    """

    files_moved: List[str] = Field([], example=["test.csv", "test2.csv"])
    options: OptionsCommon = Field(OptionsCommon(**OptionsCommon.example()), example=OptionsCommon.example())

    @staticmethod
    def example() -> dict:
        return {
            "files_moved": ["test.csv", "test2.csv"],
            "options": {"mime_type": "text/plain"},
        }


class MemoryCommon(BaseModel):
    """
    This model represents the expected memory content to communicate state of the agent.
    """

    uploads: List[int] = Field([], example=[4, 5])

    @staticmethod
    def example() -> dict:
        return {"uploads": [4, 5]}


#
# Register Models
#


class ParamsRegister(ParamsCommon):
    pass


class RequestRegister(RequestCommon):
    method: str = Field("register", const="register")
    params: ParamsRegister = Field(ParamsRegister(), const=ParamsRegister())


class ResultRegister(ResultCommon):
    errors: Optional[List[str]] = None
    logs: Optional[List[str]] = None
    memory: Optional[Dict[str, Any]] = None
    messages: Optional[List[Dict[str, Any]]] = None
    name: str = config.BasicSettings().application_name.replace(" ", "")
    display_name: str = config.BasicSettings().application_name
    description: str = config.agent_description
    default_options: Dict[str, Any] = Field(OptionsCommon(**OptionsCommon.example()), example=OptionsCommon.example())


class ResponseRegister(ResponseCommon):
    result: ResultRegister = ResultRegister()


#
# Receive Models
#


class MessageInputReceive(BaseModel):
    payload: PayloadInput


class ParamsReceive(ParamsCommon):
    message: MessageInputReceive = Field(..., example=MessageInputReceive(payload=PayloadInput()))
    options: OptionsCommon = Field(
        OptionsCommon(**OptionsCommon.example()),
        example=OptionsCommon.example(),
    )
    memory: MemoryCommon = Field(MemoryCommon(), example=MemoryCommon(**MemoryCommon.example()))
    credentials: List[CredentialsCommon] = Field([])


class RequestReceive(RequestCommon):
    method: str = Field("receive", const="receive")
    params: ParamsReceive = Field(
        ParamsReceive(message=MessageInputReceive(payload=PayloadInput(**PayloadInput.example()))),
        example=ParamsReceive(message=MessageInputReceive(payload=PayloadInput())),
    )


class ResultReceive(ResultCommon):
    errors: List[str] = Field([], example=[])
    logs: List[str] = Field([], example=[])
    memory: MemoryCommon = Field(MemoryCommon(), example=MemoryCommon(**MemoryCommon.example()))
    messages: List[MessageOutput] = Field([], example=[MessageOutput(**MessageOutput.example())])


class ResponseReceive(ResponseCommon):
    result: ResultReceive = Field(ResultReceive(), example=ResultReceive())


#
# Check models
#


class ParamsCheck(ParamsCommon):
    message: Any = Field(None, const=None)
    options: OptionsCommon = Field(
        OptionsCommon(**OptionsCommon.example()),
        example=OptionsCommon.example(),
    )
    memory: MemoryCommon = Field(MemoryCommon(), example=MemoryCommon(**MemoryCommon.example()))
    credentials: List[CredentialsCommon] = Field([])


class RequestCheck(RequestCommon):
    method: str = Field("check", const="check")
    params: ParamsCheck = Field(ParamsCheck(), example=ParamsCheck())


class ResultCheck(ResultCommon):
    errors: List[str] = Field([], example=[])
    logs: List[str] = Field([], example=[])
    memory: MemoryCommon = Field(MemoryCommon(), example=MemoryCommon(**MemoryCommon.example()))
    messages: List[MessageOutput] = Field([], example=[MessageOutput(**MessageOutput.example())])
    name: Optional[str] = None
    display_name: Optional[str] = None
    description: Optional[str] = None
    default_options: Optional[Dict[str, Any]] = None


class ResponseCheck(ResponseCommon):
    result: ResultCheck = Field(ResultCheck(), example=ResultCheck())
