# SPDX-FileCopyrightText: 2020 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import datetime

from pydantic import BaseModel, Field
from typing import List, Optional


class FileInfo(BaseModel):
    """
    See [CDSTAR Docs](https://cdstar.gwdg.de/docs/dev/#FileInfo) for more information. This model is a reduced version.
    """

    id: Optional[str] = Field(
        ...,
        example="abcdef123456",
        description="A unique and immutable string identifier. Other than the name attribute, the id will not change for the lifetime of the file and can be used to track individual files across name changes.",
    )
    name: str = Field(
        None,
        example="/example.txt",
        description="File name (unicode), always starting with a slash (/). The file name may actually represent a path and contain several path seperators (slash, /).",
    )
    type: str = Field(
        None,
        example="text/plain",
        description="User supplied or auto-detected media type. Defaults to application/octet-stream",
    )
    size: int = Field(None, example=7, description="File size in bytes")
    created: datetime.datetime = Field(
        None,
        example=datetime.datetime.fromisoformat("2020-05-02T12:15:23"),
        description="Time this file was created.",
    )
    modified: datetime.datetime = Field(
        None,
        example=datetime.datetime.fromisoformat("2020-05-03T23:11:32"),
        description="Last time the file content was modified.",
    )


class ArchiveInfo(BaseModel):
    """
    See [CDSTAR Docs](https://cdstar.gwdg.de/docs/dev/#ArchiveInfo) for more information. This model is a reduced version.
    """

    id: Optional[str] = Field(..., example="abcd1234", description="Unique ID of this archive.")
    vault: str = Field(None, example="sources", description="Name of the containing vault.")
    created: datetime.datetime = Field(
        None,
        example=datetime.datetime.fromisoformat("2020-05-02T12:15:23"),
        description="Time this archive was created.",
    )
    modified: datetime.datetime = Field(
        None,
        example=datetime.datetime.fromisoformat("2020-05-03T23:11:32"),
        description="Last time this archive, its meta-data or any of its files were modified. Note that changes to administrative meta-data (owner, ACL) do not update the modification time of an archive. If you need to track changes in administrative meta-data, always compare the actual values.",
    )
    file_count: int = Field(
        None,
        example="2",
        description="Total number of files in this archive. May be -1 to indicate that the actual number is not known. This may happen if the user does not have the permission to list the archives content.",
    )
    files: List[FileInfo] = Field(
        [],
        example=[FileInfo(id="abcdef123456"), FileInfo(id="fedcba654321")],
        description="List of files in this archive. May be incomplete or missing based on query parameters, permissions and server configuration. See Get Archive Info for details.",
    )
