# SPDX-FileCopyrightText: 2020 UMG MeDIC <marcel.parciak@med.uni-goettingen.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import re
import tarfile
from typing import List, Union

from uploader.config import BasicSettings
from uploader import models_activeworkflow as awmodels


def retrieve_files_by_mask(subdir: str, filemask: str) -> List[str]:
    """Get a List of filepaths from a directory.

    Iterate through all files in a directory, filter them with a filemask and retrieve a list of full filepaths that match the filemask.

    Paramters
    ---------
    subdir : str
        String representation of a path to a directory.
    filemask : str
        String representation of a Python regex pattern that will be used to filter files in `directory`.

    Results
    -------
    List[str]
        List of full paths as string that match `filemask` in `subdir`, including the `base_directory`.
    """
    subdir = subdir.lstrip("/")
    directory = os.path.join(BasicSettings().base_directory, subdir)
    if not os.path.exists(directory) or not os.path.isdir(directory):
        return []
    files = []
    for file in os.listdir(directory):
        if re.fullmatch(filemask, file):
            files.append(os.path.join(directory, file))
    return files


def retrieve_files_by_name(subdir: str, namelist: List[str]) -> List[str]:
    """Get a list of filepaths based on a list of filenames.

    Iterate through all files from `namelist`, check if they are available and readable from `subdir` under `base_directory`
    and retrieve the full filepath for that filename.

    Parameters
    ----------
    subdir : str
        String representation of a path to a directory
    namelist : List[str]
        List of strings representing filenames under `subdir`

    Results
    -------
    List[str]
        List of full paths as strings that are readable from `base_directory` / `subdir`
    """
    subdir = subdir.lstrip("/")
    directory = os.path.join(BasicSettings().base_directory, subdir)
    if not os.path.exists(directory) or not os.path.isdir(directory):
        return []
    files = []
    for file in namelist:
        filepath = os.path.join(directory, file)
        if os.path.isfile(filepath):
            files.append(filepath)
    return files


def tar_before_upload(files: List[str], tar_filename: str = None):
    """Compress a list of files into .tar.gz archive.

    Parameters
    ----------
    files : List[str]
        List of strings representing filenames (full path to file)
    tar_filename : str
        Designated filename for the created archive

    Results
    -------
    List[str]
        List of one containing the full path to created archive
    """
    if not tar_filename:
        dir = os.path.dirname(files[0])
        tar_filename = dir + "/archive.tar.gz"
    elif not tar_filename.endswith(".tar.gz"):
        tar_filename += ".tar.gz"
    with tarfile.open(tar_filename, "w:gz") as archive:
        for file in files:
            archive.add(file)
            os.remove(file)
    return [tar_filename]


def is_authorized(params: Union[awmodels.ParamsRegister, awmodels.ParamsReceive, awmodels.ParamsCheck]) -> bool:
    """Check if a params section is authorized to use this datalake agent instance.

    This method will check 1) if credentials are set in the configuration of the agent and 2)
    whether the supplied params are authorized to use the agent. If not credentials are set
    (which is the default), the method will return True. In any other case, the `api_key`
    option has to contain the credential name in Liquid Templating notation which refers to
    the credentials parameter.

    Parameters
    ----------
    params: Union[awmodels.ParamsRegister, awmodels.ParamsReceive, awmodels.ParamsCheck]
        Any params dictionary posted by the requester JSON

    Results
    -------
    bool
        True if parameters hold valid credentials, authorizing any request. False otherwise
    """

    if len(BasicSettings().application_api_keys) == 0:
        # If there are no credentials set, no credentials are needed.
        return True

    set_api_key = params.options.api_key
    if set_api_key.startswith("{%") and set_api_key.endswith("%}"):
        set_api_key = set_api_key[2:-2].strip()
        if set_api_key.startswith("credential"):
            search_credential_name = set_api_key[len("credential") :].strip()
        else:
            # If you end up here: an api key like `{% something %}` has been supplied. Either use
            # the `credential` key word or use the plain API key, but do not mix.
            return False
    else:
        if set_api_key.strip() in BasicSettings().application_api_keys:
            return True
        else:
            return False

    for credential in params.credentials:
        if credential.name == search_credential_name:
            if credential.value in BasicSettings().application_api_keys:
                return True

    return False
